@@ -1,4 +1,4 @@
-/*      $Id: focus.c,v 1.28 2004/10/26 20:42:22 olivier Exp $
+/*      $Id: focus.c 14694 2005-06-02 20:22:58Z olivier $
  
         This program is free software; you can redistribute it and/or modify
         it under the terms of the GNU General Public License as published by
@@ -81,10 +81,9 @@ clientGetTopMostFocusable (ScreenInfo *screen_info, int layer, Client * exclude)
         
         if (!exclude || (c != exclude))
         {
-            if ((c->win_layer <= layer)
-                 && FLAG_TEST (c->xfwm_flags, XFWM_FLAG_VISIBLE))
+            if ((c->win_layer <= layer) && FLAG_TEST (c->xfwm_flags, XFWM_FLAG_VISIBLE))
             {
-                if (clientSelectMask (c, 0, WINDOW_NORMAL | WINDOW_DIALOG | WINDOW_MODAL_DIALOG))
+                if (clientSelectMask (c, 0, WINDOW_REGULAR_FOCUSABLE))
                 {
                     top_client.prefered = c;
                 }
@@ -104,15 +103,21 @@ void
 clientFocusTop (ScreenInfo *screen_info, int layer)
 {
     ClientPair top_client;
+    DisplayInfo *display_info = NULL;
 
+    display_info = screen_info->display_info;
     top_client = clientGetTopMostFocusable (screen_info, layer, NULL);
     if (top_client.prefered)
     {
-        clientSetFocus (screen_info, top_client.prefered, GDK_CURRENT_TIME, NO_FOCUS_FLAG);
+        clientSetFocus (screen_info, top_client.prefered, 
+                        myDisplayGetCurrentTime (screen_info->display_info), 
+                        NO_FOCUS_FLAG);
     }
     else
     {
-        clientSetFocus (screen_info, top_client.highest, GDK_CURRENT_TIME, NO_FOCUS_FLAG);
+        clientSetFocus (screen_info, top_client.highest, 
+                        myDisplayGetCurrentTime (screen_info->display_info), 
+                        NO_FOCUS_FLAG);
     }
 }
 
@@ -120,6 +125,7 @@ void
 clientFocusNew(Client * c)
 {
     ScreenInfo *screen_info = NULL;
+    DisplayInfo *display_info = NULL;
     gboolean give_focus;
     gboolean prevent_focus_stealing;
 
@@ -130,6 +136,7 @@ clientFocusNew(Client * c)
         return;
     }
     screen_info = c->screen_info;
+    display_info = screen_info->display_info;
     give_focus = screen_info->params->focus_new;
     prevent_focus_stealing = screen_info->params->prevent_focus_stealing;
 
@@ -147,6 +154,7 @@ clientFocusNew(Client * c)
         }
     }
 
+    clientGrabMouseButton (c);
     if (give_focus || FLAG_TEST(c->flags, CLIENT_FLAG_STATE_MODAL))
     {
         if (client_focus)
@@ -156,14 +164,18 @@ clientFocusNew(Client * c)
                 clientRaise (c);
             }
         }
-        clientSetFocus (c->screen_info, c, GDK_CURRENT_TIME, FOCUS_IGNORE_MODAL);
-        clientPassGrabButton1 (c);
+        clientSetFocus (screen_info, c, 
+#if 0
+                        myDisplayGetCurrentTime (display_info), 
+#else
+                        CurrentTime, 
+#endif
+                        FOCUS_IGNORE_MODAL);
     }
     else
     {
         FLAG_SET (c->flags, CLIENT_FLAG_DEMANDS_ATTENTION);
         clientSetNetState (c);
-        clientPassGrabButton1 (NULL);
     }
 }
 
@@ -200,6 +212,7 @@ clientSelectMask (Client * c, int mask, int type)
     {
         return TRUE;
     }
+
     return FALSE;
 }
 
@@ -217,7 +230,7 @@ clientGetNext (Client * c, int mask)
         for (c2 = c->next, i = 0; (c2) && (i < screen_info->client_count - 1);
             c2 = c2->next, i++)
         {
-            if (clientSelectMask (c2, mask, WINDOW_NORMAL | WINDOW_DIALOG | WINDOW_MODAL_DIALOG))
+            if (clientSelectMask (c2, mask, WINDOW_REGULAR_FOCUSABLE))
             {
                 return c2;
             }
@@ -240,7 +253,7 @@ clientGetPrevious (Client * c, int mask)
         for (c2 = c->prev, i = 0; (c2) && (i < screen_info->client_count);
             c2 = c2->prev, i++)
         {
-            if (clientSelectMask (c2, mask, WINDOW_NORMAL | WINDOW_DIALOG | WINDOW_MODAL_DIALOG))
+            if (clientSelectMask (c2, mask, WINDOW_REGULAR_FOCUSABLE))
             {
                 return c2;
             }
@@ -252,6 +265,7 @@ clientGetPrevious (Client * c, int mask)
 void
 clientPassFocus (ScreenInfo *screen_info, Client *c, Client *exclude)
 {
+    DisplayInfo *display_info = NULL;
     Client *new_focus = NULL;
     Client *current_focus = client_focus;
     ClientPair top_most;
@@ -273,6 +287,12 @@ clientPassFocus (ScreenInfo *screen_info, Client *c, Client *exclude)
         return;
     }
 
+    if (current_focus == last_ungrab)
+    {
+        clientPassGrabMouseButton (NULL);
+    }
+
+    display_info = screen_info->display_info;
     top_most = clientGetTopMostFocusable (screen_info, look_in_layer, exclude);
     if (screen_info->params->click_to_focus)
     {
@@ -285,7 +305,6 @@ clientPassFocus (ScreenInfo *screen_info, Client *c, Client *exclude)
                    for group, so it safe to use clientGetTransient because 
                    it's really what we want...
                  */
-
                 c2 = clientGetTransient (c);
                 if (c2 && FLAG_TEST(c2->xfwm_flags, XFWM_FLAG_VISIBLE))
                 {
@@ -294,6 +313,14 @@ clientPassFocus (ScreenInfo *screen_info, Client *c, Client *exclude)
                     clientRaise (c2);
                 }
             }
+            else
+            {
+                c2 = clientGetNext (c, 0);
+                if ((c2) && (c2->win_layer >= c->win_layer))
+                {
+                    new_focus = c2;
+                }
+            }
         }
     }
     else if (XQueryPointer (myScreenGetXDisplay (screen_info), screen_info->xroot, &dr, &window, &rx, &ry, &wx, &wy, &mask))
@@ -304,15 +331,13 @@ clientPassFocus (ScreenInfo *screen_info, Client *c, Client *exclude)
     {
         new_focus = top_most.prefered ? top_most.prefered : top_most.highest;
     }
-    clientSetFocus (screen_info, new_focus, GDK_CURRENT_TIME, FOCUS_IGNORE_MODAL | FOCUS_FORCE);
-    if (new_focus == top_most.highest)
-    {
-        clientPassGrabButton1 (new_focus);
-    }
-    else if (last_ungrab == c)
-    {
-        clientPassGrabButton1 (NULL);
-    }
+    clientSetFocus (screen_info, new_focus, 
+#if 0
+                    myDisplayGetCurrentTime (screen_info->display_info), 
+#else
+                    CurrentTime, 
+#endif
+                    FOCUS_IGNORE_MODAL | FOCUS_FORCE);
 }
 
 gboolean
@@ -365,22 +390,35 @@ clientSortRing(Client *c)
 void
 clientUpdateFocus (ScreenInfo *screen_info, Client * c, unsigned short flags)
 {
+    DisplayInfo *display_info = NULL;
     Client *c2 = ((client_focus != c) ? client_focus : NULL);
     unsigned long data[2];
 
     TRACE ("entering clientUpdateFocus");
 
+    display_info = screen_info->display_info;
     pending_focus = NULL;
     if ((c) && !clientAcceptFocus (c))
     {
         TRACE ("SKIP_FOCUS set for client \"%s\" (0x%lx)", c->name, c->window);
         return;
     }
+
     if ((c) && (c == client_focus) && !(flags & FOCUS_FORCE))
     {
         TRACE ("client \"%s\" (0x%lx) is already focused, ignoring request", c->name, c->window);
         return;
     }
+
+    /* 
+       We can release the button mouse grab if we don't raise on click or if the focused window
+       is the one that has been raised at last.
+     */
+    if (!(screen_info->params->raise_on_click) || (c == clientGetLastRaise (screen_info)))
+    {
+        clientPassGrabMouseButton (c);
+    }
+
     client_focus = c;
     if (c)
     {
@@ -402,23 +440,25 @@ clientUpdateFocus (ScreenInfo *screen_info, Client * c, unsigned short flags)
         if (c)
         {
             clientAdjustFullscreenLayer (c2, FALSE);
-            clientRaise (c);
-            clientPassGrabButton1 (c);
+            /* clientRaise (c); */
         }
         frameDraw (c2, FALSE, FALSE);
     }
     data[1] = None;
-    XChangeProperty (myScreenGetXDisplay (screen_info), screen_info->xroot, net_active_window, XA_WINDOW, 32,
-        PropModeReplace, (unsigned char *) data, 2);
+    XChangeProperty (display_info->dpy, screen_info->xroot, 
+                     display_info->atoms[NET_ACTIVE_WINDOW], XA_WINDOW, 32,
+                     PropModeReplace, (unsigned char *) data, 2);
 }
 
 void
 clientSetFocus (ScreenInfo *screen_info, Client * c, Time timestamp, unsigned short flags)
 {
+    DisplayInfo *display_info = NULL;
     Client *c2 = NULL;
 
     TRACE ("entering clientSetFocus");
     
+    display_info = screen_info->display_info;
     if ((c) && !(flags & FOCUS_IGNORE_MODAL))
     {
         c2 = clientGetModalFor (c);
@@ -450,7 +490,7 @@ clientSetFocus (ScreenInfo *screen_info, Client * c, Time timestamp, unsigned sh
         }
         if (FLAG_TEST(c->wm_flags, WM_FLAG_TAKEFOCUS))
         {
-            sendClientMessage (c->screen_info, c->window, wm_protocols, wm_takefocus, timestamp);
+            sendClientMessage (c->screen_info, c->window, WM_TAKE_FOCUS, timestamp);
         }
         if (FLAG_TEST(c->flags, CLIENT_FLAG_DEMANDS_ATTENTION))
         {
@@ -469,12 +509,12 @@ clientSetFocus (ScreenInfo *screen_info, Client * c, Time timestamp, unsigned sh
         if (c2)
         {
             frameDraw (c2, FALSE, FALSE);
-            XChangeProperty (clientGetXDisplay (c2), c2->screen_info->xroot, net_active_window, XA_WINDOW, 32,
+            XChangeProperty (clientGetXDisplay (c2), c2->screen_info->xroot, display_info->atoms[NET_ACTIVE_WINDOW], XA_WINDOW, 32,
                              PropModeReplace, (unsigned char *) data, 2);
         }
-        XChangeProperty (myScreenGetXDisplay (screen_info), screen_info->xroot, net_active_window, XA_WINDOW, 32,
+        XChangeProperty (myScreenGetXDisplay (screen_info), screen_info->xroot, display_info->atoms[NET_ACTIVE_WINDOW], XA_WINDOW, 32,
                          PropModeReplace, (unsigned char *) data, 2);
-        XSetInputFocus (myScreenGetXDisplay (screen_info), screen_info->gnome_win, RevertToPointerRoot, GDK_CURRENT_TIME);
+        XSetInputFocus (myScreenGetXDisplay (screen_info), screen_info->gnome_win, RevertToPointerRoot, timestamp);
     }
 }
 
@@ -491,63 +531,115 @@ clientClearFocus (void)
 }
 
 void
-clientGrabButton1 (Client * c)
+clientGrabMouseButton (Client * c)
 {
-    g_return_if_fail (c != NULL);
-    TRACE ("entering clientGrabButton1");
-    TRACE ("grabbing buttons for client \"%s\" (0x%lx)", c->name, c->window);
+    ScreenInfo *screen_info = NULL;
     
-    grabButton(clientGetXDisplay (c), Button1, 0, c->window);
+    g_return_if_fail (c != NULL);
+    TRACE ("entering clientGrabMouseButton");
+    TRACE ("grabbing buttons for client \"%s\" (0x%lx)", c->name, c->window);
+
+    screen_info = c->screen_info;
+    if (screen_info->params->raise_with_any_button)
+    {
+        grabButton(clientGetXDisplay (c), AnyButton, 0, c->window);
+    }
+    else
+    {
+        grabButton(clientGetXDisplay (c), Button1, 0, c->window);
+    }
 }
 
 void
-clientUngrabButton1 (Client * c)
+clientUngrabMouseButton (Client * c)
 {
+    ScreenInfo *screen_info = NULL;
+    
     g_return_if_fail (c != NULL);
-    TRACE ("entering clientUngrabButton1");
+    TRACE ("entering clientUngrabMouseButton");
     TRACE ("ungrabing buttons for client \"%s\" (0x%lx)", c->name, c->window);
 
-    ungrabButton(clientGetXDisplay (c), Button1, 0, c->window);
+    screen_info = c->screen_info;
+    if (screen_info->params->raise_with_any_button)
+    {
+        ungrabButton(clientGetXDisplay (c), AnyButton, 0, c->window);
+    }
+    else
+    {
+        ungrabButton(clientGetXDisplay (c), Button1, 0, c->window);
+    }
 }
 
 void
-clientPassGrabButton1(Client * c)
+clientGrabMouseButtonForAll (ScreenInfo *screen_info)
 {
-    TRACE ("entering clientPassGrabButton1");
+    Client *c;
+    int i;
+    
+    g_return_if_fail (screen_info != NULL);
+    TRACE ("entering clientGrabMouseButtonForAll");
+
+    for (c = screen_info->clients, i = 0; (c) && (i < screen_info->client_count); c = c->next, i++)
+    {
+        clientGrabMouseButton (c);
+    }
+    clientClearLastUngrab ();
+}
+
+void
+clientUngrabMouseButtonForAll (ScreenInfo *screen_info)
+{
+    Client *c;
+    int i;
+    
+    g_return_if_fail (screen_info != NULL);
+    TRACE ("entering clientUngrabMouseButtonForAll");
+
+    for (c = screen_info->clients, i = 0; (c) && (i < screen_info->client_count); c = c->next, i++)
+    {
+        clientUngrabMouseButton (c);
+    }
+    clientClearLastUngrab ();
+}
+
+void
+clientPassGrabMouseButton (Client * c)
+{
+    TRACE ("entering clientPassMouseGrabButton");
     TRACE ("ungrabing buttons for client \"%s\" (0x%lx)", c->name, c->window);
 
     if (c == NULL)
     {
         if (last_ungrab)
         {
-            clientGrabButton1 (last_ungrab);
+            clientGrabMouseButton (last_ungrab);
         }
         last_ungrab = NULL;
         return;
     }
-    
+
     if (last_ungrab == c)
     {
         return;
     }
-    
+
     if (last_ungrab)
     {
-        clientGrabButton1 (last_ungrab);
+        clientGrabMouseButton (last_ungrab);
     }
-    
-    clientUngrabButton1 (c);
+
+    clientUngrabMouseButton (c);
     last_ungrab = c;
 }
 
 Client *
-clientGetLastGrab (void)
+clientGetLastUngrab (void)
 {
     return last_ungrab;
 }
 
 void
-clientClearLastGrab (void)
+clientClearLastUngrab (void)
 {
     last_ungrab = NULL;
 }
